from django.urls import reverse
from django.utils import timezone

from furl import furl
from onelogin.saml2.settings import OneLogin_Saml2_Settings

from .generate_digid_metadata import SamlMetadataBaseCommand


class Command(SamlMetadataBaseCommand):
    help = "Create the eHerkenning metadata file"

    required_options = [
        "key_file",
        "cert_file",
        "entity_id",
        "base_url",
        "service_name",
        "service_description",
        "oin",
    ]

    def add_arguments(self, parser):
        super().add_arguments(parser)
        parser.add_argument(
            "--loa",
            type=str,
            action="store",
            help="Level of Assurance (LoA) to use for all the services.",
            default="urn:etoegang:core:assurance-class:loa3",
        )
        parser.add_argument(
            "--eh_attribute_consuming_service_index",
            type=str,
            action="store",
            help="Attribute consuming service index for the eHerkenning service, defaults to 9052",
            default="9052",
        )
        parser.add_argument(
            "--eidas_attribute_consuming_service_index",
            type=str,
            action="store",
            help="Attribute consuming service index for the eHerkenning service, defaults to 9053",
            default="9053",
        )
        parser.add_argument(
            "--oin",
            type=str,
            action="store",
            help="The OIN of the company providing the service.",
            default=None,
        )
        parser.add_argument(
            "--no_eidas",
            action="store_true",
            help="If True, then the service catalogue will contain only the eHerkenning service. Defaults to False.",
            default=False,
        )

    def get_filename(self):
        date_string = timezone.now().date().isoformat()
        return f"eherkenning-metadata-{date_string}.xml"

    def generate_metadata(self, options):
        setting_dict = {
            "strict": True,
            "security": {
                "signMetadata": True,
                "authnRequestsSigned": True,
                "wantAssertionsEncrypted": options["want_assertions_encrypted"],
                "wantAssertionsSigned": options["want_assertions_signed"],
                "soapClientKey": options["key_file"],
                "soapClientCert": options["cert_file"],
                "soapClientPassphrase": options["key_passphrase"],
                "signatureAlgorithm": options["signature_algorithm"],
                "digestAlgorithm": options["digest_algorithm"],
                # See comment in the python3-saml for in  OneLogin_Saml2_Response.validate_num_assertions (onelogin/saml2/response.py)
                # for why we need this option.
                "disableSignatureWrappingProtection": True,
                # For eHerkenning, if the Metadata file expires, we sent them an update. So
                # there is no need for an expiry date.
                "metadataValidUntil": "",
                "metadataCacheDuration": "",
                "requestedAuthnContextComparison": "minimum",
                "requestedAuthnContext": [options["loa"]],
            },
            # Service Provider Data that we are deploying.
            "sp": {
                # Identifier of the SP entity  (must be a URI)
                "entityId": options["entity_id"],
                # Specifies info about where and how the <AuthnResponse> message MUST be
                # returned to the requester, in this case our SP.
                "assertionConsumerService": {
                    "url": furl(options["base_url"] + reverse("eherkenning:acs")).url,
                    "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Artifact",
                },
                "attributeConsumingServices": [
                    {
                        "index": options["eh_attribute_consuming_service_index"],
                        "serviceName": options["service_name"],
                        "serviceDescription": options["service_description"],
                        "requestedAttributes": [
                            {
                                "name": "urn:etoegang:DV:%(oin)s:services:%(index)s"
                                % {
                                    "oin": options["oin"],
                                    "index": options[
                                        "eh_attribute_consuming_service_index"
                                    ],
                                },
                                "isRequired": False,
                            }
                        ],
                        "language": "nl",
                    },
                    {
                        "index": options["eidas_attribute_consuming_service_index"],
                        "serviceName": options["service_name"] + " (eIDAS)",
                        "serviceDescription": options["service_description"],
                        "requestedAttributes": [
                            {
                                "name": "urn:etoegang:DV:%(oin)s:services:%(index)s"
                                % {
                                    "oin": options["oin"],
                                    "index": options[
                                        "eidas_attribute_consuming_service_index"
                                    ],
                                },
                                "isRequired": False,
                            }
                        ],
                        "language": "nl",
                    },
                ],
                "NameIDFormat": "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified",
                "x509cert": open(options["cert_file"], "r").read(),
                "privateKey": open(options["key_file"], "r").read(),
                "privateKeyPassphrase": options["key_passphrase"],
            },
        }

        telephone = options["technical_contact_person_telephone"]
        email = options["technical_contact_person_email"]
        if telephone and email:
            setting_dict["contactPerson"] = {
                "technical": {"telephoneNumber": telephone, "emailAddress": email}
            }

        if options["organization_url"] and options["organization_name"]:
            setting_dict["organization"] = {
                "nl": {
                    "name": options["organization_name"],
                    "displayname": options["organization_name"],
                    "url": options["organization_url"],
                }
            }

        if options["no_eidas"]:
            setting_dict["sp"]["attributeConsumingServices"] = setting_dict["sp"][
                "attributeConsumingServices"
            ][:-1]

        saml2_settings = OneLogin_Saml2_Settings(setting_dict, sp_validation_only=True)
        return saml2_settings.get_sp_metadata()



from django.urls import path

from . import views

app_name = "digid"

urlpatterns = [
    path("login/", views.DigiDLoginView.as_view(), name="login"),
    path("acs/", views.DigiDAssertionConsumerServiceView.as_view(), name="acs"),
]



from django.core.management.base import BaseCommand, CommandError
from django.utils.module_loading import import_string


class Command(BaseCommand):
    help = "Create the various SAML metadata files."

    def add_arguments(self, parser):
        parser.add_argument(
            "-c",
            "--classes",
            dest="client_classes",
            default=[
                "digid_eherkenning.saml2.digid.DigiDClient",
                "digid_eherkenning.saml2.eherkenning.eHerkenningClient",
            ],
            nargs="*",
            type=str,
            help="SAML2 client class to generate metadata for",
        )

    def handle(self, *args, **options):
        client_classes = [import_string(c) for c in options.get("client_classes")]

        for ClientClass in client_classes:
            client = ClientClass()
            try:
                client.write_metadata()
            except FileExistsError as e:
                raise CommandError(f"The file {e.filename} already exists.")



from django.conf.urls import url

from . import views

app_name = "eherkenning"

urlpatterns = [
    url(r"login/", views.eHerkenningLoginView.as_view(), name="login"),
    url(r"acs/", views.eHerkenningAssertionConsumerServiceView.as_view(), name="acs"),
]



from django.conf import settings
from django.urls import reverse

from .base import BaseSaml2Client


class DigiDClient(BaseSaml2Client):
    cache_key_prefix = "digid"
    cache_timeout = 60 * 60  # 1 hour

    def __init__(self):
        conf = settings.DIGID.copy()
        conf.setdefault("acs_path", reverse("digid:acs"))

        super().__init__(conf)

    def create_config(self, config_dict):
        config_dict["security"].update(
            {
                # None sent for digi-id.
                "wantAttributeStatement": False,
                # For DigiD, if the Metadata file expires, we sent them an update. So
                # there is no need for an expiry date.
                "metadataValidUntil": "",
                "metadataCacheDuration": "",
                "requestedAuthnContextComparison": "minimum",
                "requestedAuthnContext": [
                    "urn:oasis:names:tc:SAML:2.0:ac:classes:MobileTwoFactorContract",
                ],
            }
        )
        return super().create_config(config_dict)

    def create_authn_request(self, request, return_to=None):
        return super().create_authn_request(
            request,
            return_to=return_to,
            is_passive=False,
            set_nameid_policy=False,
            name_id_value_req=None,
        )



from django.core.management.base import BaseCommand, CommandError
from django.urls import reverse
from django.utils import timezone

from onelogin.saml2.settings import OneLogin_Saml2_Settings


class SamlMetadataBaseCommand(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "--want_assertions_encrypted",
            action="store_true",
            help="If True the XML assertions need to be encrypted. Defaults to False",
            default=False,
        )
        parser.add_argument(
            "--want_assertions_signed",
            action="store_true",
            help="If True, the XML assertions need to be signed, otherwise the whole response needs to be signed. Defaults to True",
            default=True,
        )
        parser.add_argument(
            "--key_file",
            action="store",
            type=str,
            help="The filepath to the TLS key. This will be used both by the SOAP client and for signing the requests.",
            default=None,
        )
        parser.add_argument(
            "--cert_file",
            action="store",
            type=str,
            help="The filepath to the TLS certificate. This will be used both by the SOAP client and for signing the requests.",
            default=None,
        )
        parser.add_argument(
            "--key_passphrase",
            type=str,
            action="store",
            help="Passphrase for SOAP client",
            default=None,
        )
        parser.add_argument(
            "--signature_algorithm",
            type=str,
            action="store",
            help="Signature algorithm, defaults to RSA_SHA1",
            default="http://www.w3.org/2000/09/xmldsig#rsa-sha1",
        )

        parser.add_argument(
            "--digest_algorithm",
            type=str,
            action="store",
            help="Digest algorithm, defaults to SHA1",
            default="http://www.w3.org/2000/09/xmldsig#sha1",
        )
        parser.add_argument(
            "--entity_id",
            type=str,
            action="store",
            help="Service provider entity ID",
            default=None,
        )
        parser.add_argument(
            "--base_url",
            type=str,
            action="store",
            help="Base URL of the application",
            default=None,
        )
        parser.add_argument(
            "--service_name",
            type=str,
            action="store",
            help="The name of the service for which DigiD login is required",
            default=None,
        )
        parser.add_argument(
            "--service_description",
            type=str,
            action="store",
            help="A description of the service for which DigiD login is required",
            default=None,
        )
        parser.add_argument(
            "--technical_contact_person_telephone",
            type=str,
            action="store",
            help="Telephone number of the technical person responsible for this DigiD setup. For it to be used, --technical_contact_person_email should also be set.",
            default=None,
        )
        parser.add_argument(
            "--technical_contact_person_email",
            type=str,
            action="store",
            help="Email address of the technical person responsible for this DigiD setup. For it to be used, --technical_contact_person_telephone should also be set.",
            default=None,
        )
        parser.add_argument(
            "--organization_name",
            type=str,
            action="store",
            help="Name of the organisation providing the service for which DigiD login is setup. For it to be used, also --organization_url should be filled.",
            default=None,
        )
        parser.add_argument(
            "--organization_url",
            type=str,
            action="store",
            help="URL of the organisation providing the service for which DigiD login is setup. For it to be used, also --organization_name should be filled.",
            default=None,
        )
        parser.add_argument(
            "--output_file",
            type=str,
            action="store",
            help="Name of the file to which to write the metadata. Otherwise will be printed on stdout",
            default=None,
        )
        parser.add_argument(
            "--test",
            action="store_true",
            help="If True the metadata is printed to stdout. Defaults to False",
            default=False,
        )

    def check_options(self, options: dict) -> None:
        missing_required_options = []
        for option in self.required_options:
            if option not in options or not options[option]:
                missing_required_options.append(option)

        if len(missing_required_options) > 0:
            message = "Missing the following required arguments: %s" % " ".join(
                [f"--{option}" for option in missing_required_options]
            )
            raise CommandError(message)

    def handle(self, *args, **options):
        self.check_options(options)

        metadata_content = self.generate_metadata(options)

        if options["test"]:
            self.stdout.write(metadata_content.decode("utf-8"))
            return

        if options["output_file"]:
            filename = options["output_file"]
        else:
            filename = self.get_filename()

        metadata_file = open(filename, "xb")
        metadata_file.write(metadata_content)
        metadata_file.close()

        self.stdout.write(
            self.style.SUCCESS("Metadata file successfully generated: %s" % filename)
        )


class Command(SamlMetadataBaseCommand):
    help = "Create the DigiD metadata file"

    required_options = [
        "key_file",
        "cert_file",
        "entity_id",
        "base_url",
        "service_name",
        "service_description",
    ]

    def add_arguments(self, parser):
        super().add_arguments(parser)

        parser.add_argument(
            "--attribute_consuming_service_index",
            type=str,
            action="store",
            help="Attribute consuming service index, defaults to 1",
            default="1",
        )

    def get_filename(self):
        date_string = timezone.now().date().isoformat()
        return f"digid-metadata-{date_string}.xml"

    def generate_metadata(self, options):
        setting_dict = {
            "strict": True,
            "security": {
                "signMetadata": True,
                "authnRequestsSigned": True,
                "wantAssertionsEncrypted": options["want_assertions_encrypted"],
                "wantAssertionsSigned": options["want_assertions_signed"],
                # None sent for digi-id.
                "wantAttributeStatement": False,
                "soapClientKey": options["key_file"],
                "soapClientCert": options["cert_file"],
                "soapClientPassphrase": options["key_passphrase"],
                "signatureAlgorithm": options["signature_algorithm"],
                "digestAlgorithm": options["digest_algorithm"],
                # For DigiD, if the Metadata file expires, we sent them an update. So
                # there is no need for an expiry date.
                "metadataValidUntil": "",
                "metadataCacheDuration": "",
                "requestedAuthnContextComparison": "minimum",
                "requestedAuthnContext": [
                    "urn:oasis:names:tc:SAML:2.0:ac:classes:MobileTwoFactorContract",
                ],
            },
            # Service Provider Data that we are deploying.
            "sp": {
                # Identifier of the SP entity  (must be a URI)
                "entityId": options["entity_id"],
                # Specifies info about where and how the <AuthnResponse> message MUST be
                # returned to the requester, in this case our SP.
                "assertionConsumerService": {
                    "url": options["base_url"] + reverse("digid:acs"),
                    "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Artifact",
                },
                # If you need to specify requested attributes, set a
                # attributeConsumingService per service. nameFormat, attributeValue and
                # friendlyName can be omitted
                "attributeConsumingService": {
                    "index": options["attribute_consuming_service_index"],
                    "serviceName": options["service_name"],
                    "serviceDescription": options["service_description"],
                    "requestedAttributes": [
                        {
                            "name": "bsn",
                            "required": True,
                        }
                    ],
                },
                "NameIDFormat": "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified",
                "x509cert": open(options["cert_file"], "r").read(),
                "privateKey": open(options["key_file"], "r").read(),
                "privateKeyPassphrase": options["key_passphrase"],
            },
        }

        telephone = options["technical_contact_person_telephone"]
        email = options["technical_contact_person_email"]
        if telephone and email:
            setting_dict["contactPerson"] = {
                "technical": {"telephoneNumber": telephone, "emailAddress": email}
            }

        if options["organization_url"] and options["organization_name"]:
            setting_dict["organization"] = {
                "nl": {
                    "name": options["organization_name"],
                    "displayname": options["organization_name"],
                    "url": options["organization_url"],
                }
            }

        saml2_settings = OneLogin_Saml2_Settings(setting_dict, sp_validation_only=True)
        return saml2_settings.get_sp_metadata()



from uuid import uuid4

from django.core.management.base import BaseCommand, CommandError
from django.utils import timezone

from digid_eherkenning.saml2.eherkenning import create_service_catalogus


def generate_dienst_catalogus(options):
    settings = {
        "key_file": options["key_file"],
        "cert_file": options["cert_file"],
        "base_url": options["base_url"],
        "entity_id": options["entity_id"],
        "oin": options["oin"],
        "organisation_name": options["organization_name"],
        "services": [
            {
                "attribute_consuming_service_index": options[
                    "eh_attribute_consuming_service_index"
                ],
                "service_loa": options["loa"],
                "service_uuid": str(uuid4()),
                "service_name": {
                    "nl": options["service_name"],
                    "en": options["service_name"],
                },
                "service_description": {
                    "nl": options["service_description"],
                    "en": options["service_description"],
                },
                "service_instance_uuid": str(uuid4()),
                "service_url": options["base_url"],
                # Either require and return RSIN and KVKNr (set 1) or require only KvKnr (set 2). The
                # latter is needed for 'eenmanszaak'
                "entity_concerned_types_allowed": [
                    {
                        "set_number": "1",
                        "name": "urn:etoegang:1.9:EntityConcernedID:RSIN",
                    },
                    {
                        "set_number": "1",
                        "name": "urn:etoegang:1.9:EntityConcernedID:KvKnr",
                    },
                    {
                        "set_number": "2",
                        "name": "urn:etoegang:1.9:EntityConcernedID:KvKnr",
                    },
                ],
                "requested_attributes": [
                    {
                        "name": "urn:etoegang:DV:%(oin)s:services:%(index)s"
                        % {
                            "oin": options["oin"],
                            "index": options["eh_attribute_consuming_service_index"],
                        },
                        "isRequired": False,
                    }
                ],
                "privacy_policy_url": {
                    "nl": options["privacy_policy"],
                },
                "herkenningsmakelaars_id": options["makelaar_id"],
            },
            {
                "attribute_consuming_service_index": options[
                    "eidas_attribute_consuming_service_index"
                ],
                "service_loa": options["loa"],
                "service_uuid": str(uuid4()),
                "service_name": {
                    "nl": options["service_name"] + " (eIDAS)",
                    "en": options["service_name"] + " (eIDAS)",
                },
                "service_description": {
                    "nl": options["service_description"],
                    "en": options["service_description"],
                },
                "service_instance_uuid": str(uuid4()),
                "service_url": options["base_url"],
                "entity_concerned_types_allowed": [
                    {
                        "set_number": "1",
                        "name": "urn:etoegang:1.9:EntityConcernedID:Pseudo",
                    },
                ],
                "requestedAttributes": [
                    {
                        "name": "urn:etoegang:DV:%(oin)s:services:%(index)s"
                        % {
                            "oin": options["oin"],
                            "index": options["eidas_attribute_consuming_service_index"],
                        },
                        "isRequired": False,
                    }
                ],
                "privacy_policy_url": {
                    "nl": options["privacy_policy"],
                },
                "herkenningsmakelaars_id": options["makelaar_id"],
                "classifiers": ["eIDAS-inbound"],
            },
        ],
    }

    if options["no_eidas"]:
        settings["services"] = settings["services"][:-1]

    return create_service_catalogus(settings)


class Command(BaseCommand):
    help = "Create the eHerkenning dienstcatalogus file"

    def add_arguments(self, parser):
        parser.add_argument(
            "--key_file",
            action="store",
            type=str,
            help="The filepath to the TLS key. This will be used both by the SOAP client and for signing the requests.",
            default=None,
        )
        parser.add_argument(
            "--cert_file",
            action="store",
            type=str,
            help="The filepath to the TLS certificate. This will be used both by the SOAP client and for signing the requests.",
            default=None,
        )
        parser.add_argument(
            "--signature_algorithm",
            type=str,
            action="store",
            help="Signature algorithm, defaults to RSA_SHA1",
            default="http://www.w3.org/2000/09/xmldsig#rsa-sha1",
        )
        parser.add_argument(
            "--loa",
            type=str,
            action="store",
            help="Level of Assurance (LoA) to use for all the services.",
            default="urn:etoegang:core:assurance-class:loa3",
        )
        parser.add_argument(
            "--digest_algorithm",
            type=str,
            action="store",
            help="Digest algorithm, defaults to SHA1",
            default="http://www.w3.org/2000/09/xmldsig#sha1",
        )
        parser.add_argument(
            "--entity_id",
            type=str,
            action="store",
            help="Service provider entity ID",
            default=None,
        )
        parser.add_argument(
            "--organization_name",
            type=str,
            action="store",
            help="Name of the organisation providing the service for which eHerkenning login is setup.",
            default=None,
        )
        parser.add_argument(
            "--base_url",
            type=str,
            action="store",
            help="Base URL of the application",
            default=None,
        )
        parser.add_argument(
            "--service_name",
            type=str,
            action="store",
            help="The name of the service for which DigiD login is required",
            default=None,
        )
        parser.add_argument(
            "--service_description",
            type=str,
            action="store",
            help="A description of the service for which DigiD login is required",
            default=None,
        )
        parser.add_argument(
            "--eh_attribute_consuming_service_index",
            type=str,
            action="store",
            help="Attribute consuming service index for the eHerkenning service, defaults to 9052",
            default="9052",
        )
        parser.add_argument(
            "--no_eidas",
            action="store_true",
            help="If True, then the service catalogue will contain only the eHerkenning service. Defaults to False.",
            default=False,
        )
        parser.add_argument(
            "--eidas_attribute_consuming_service_index",
            type=str,
            action="store",
            help="Attribute consuming service index for the eHerkenning service, defaults to 9053",
            default="9053",
        )
        parser.add_argument(
            "--oin",
            type=str,
            action="store",
            help="The OIN of the company providing the service.",
            default=None,
        )
        parser.add_argument(
            "--privacy_policy",
            type=str,
            action="store",
            help="The URL where the privacy policy from the organisation providing the service can be found.",
            default=None,
        )
        parser.add_argument(
            "--makelaar_id",
            type=str,
            action="store",
            help="OIN of the broker used to set up eHerkenning/eIDAS.",
            default=None,
        )
        parser.add_argument(
            "--output_file",
            type=str,
            action="store",
            help="Name of the file to which to write the metadata. Otherwise will be printed on stdout",
            default=None,
        )
        parser.add_argument(
            "--test",
            action="store_true",
            help="If True the metadata is printed to stdout. Defaults to False",
            default=False,
        )

    def check_options(self, options: dict) -> None:
        required_options = [
            "key_file",
            "cert_file",
            "entity_id",
            "base_url",
            "service_name",
            "service_description",
            "oin",
            "privacy_policy",
            "makelaar_id",
        ]
        missing_required_options = []
        for option in required_options:
            if option not in options or not options[option]:
                missing_required_options.append(option)

        if len(missing_required_options) > 0:
            message = "Missing the following required arguments: %s" % " ".join(
                [f"--{option}" for option in missing_required_options]
            )
            raise CommandError(message)

    def handle(self, *args, **options):
        self.check_options(options)

        catalogus = generate_dienst_catalogus(options)

        if options["test"]:
            self.stdout.write(catalogus.decode("utf-8"))
            return

        if options["output_file"]:
            filename = options["output_file"]
        else:
            date_string = timezone.now().date().isoformat()
            filename = f"eherkenning-dienstcatalogus-{date_string}.xml"

        catalogus_file = open(filename, "xb")
        catalogus_file.write(catalogus)
        catalogus_file.close()

        self.stdout.write(
            self.style.SUCCESS(
                "Dienstcatalogus file successfully generated: %s" % filename
            )
        )



import binascii
from base64 import b64encode
from io import BytesIO
from typing import List
from uuid import uuid4

from django.conf import settings
from django.urls import reverse
from django.utils import timezone

from defusedxml.lxml import tostring
from lxml.builder import ElementMaker
from lxml.etree import Element
from OpenSSL import crypto

from ..settings import EHERKENNING_DS_XSD
from ..utils import validate_xml
from .base import BaseSaml2Client, get_service_description, get_service_name

namespaces = {
    "xs": "http://www.w3.org/2001/XMLSchema",
    "md": "urn:oasis:names:tc:SAML:2.0:metadata",
    "ds": "http://www.w3.org/2000/09/xmldsig#",
    "esc": "urn:etoegang:1.13:service-catalog",
    "saml": "urn:oasis:names:tc:SAML:2.0:assertion",
}
ESC = ElementMaker(namespace=namespaces["esc"], nsmap=namespaces)
DS = ElementMaker(namespace=namespaces["ds"], nsmap=namespaces)
SAML = ElementMaker(namespace=namespaces["saml"], nsmap=namespaces)
MD = ElementMaker(namespace=namespaces["md"], nsmap=namespaces)

xml_nl_lang = {"{http://www.w3.org/XML/1998/namespace}lang": "nl"}


def xml_datetime(d):
    return d.isoformat(timespec="seconds")


def create_language_elements(element_name, option_value, default_language="en"):
    """
    Convert a configuration option into zero or more eHerkenning dienstcatalogus
    elements

    :param element_name Name of the XML element to be generated
    :param option_value Configuration option being either a string or a dictionary
                        containing the language code as key, and the option as value.
    :return list of etree elements
    """

    if option_value is None:
        options = []

    options = (
        option_value
        if isinstance(option_value, dict)
        else {default_language: option_value}
    )

    elements = []
    for lang, option in options.items():
        xml_lang = {"{http://www.w3.org/XML/1998/namespace}lang": lang}
        elements.append(
            ESC(element_name, option, **xml_lang),
        )
    return elements


def create_service_catalogue(id, issue_instant, signature, service_provider):
    ns = namespaces["esc"]
    args = [
        signature,
        service_provider,
    ]
    kwargs = {
        "ID": id,
        f"{{{ns}}}IssueInstant": xml_datetime(issue_instant),
        f"{{{ns}}}Version": "urn:etoegang:1.10:53",
    }
    return ESC("ServiceCatalogue", *args, **kwargs)


def create_signature(id):
    """
    https://afsprakenstelsel.etoegang.nl/display/as/Digital+signature
    """
    transforms = [
        DS(
            "Transform",
            Algorithm="http://www.w3.org/2000/09/xmldsig#enveloped-signature",
        ),
        DS("Transform", Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#"),
    ]

    args = [
        DS(
            "SignedInfo",
            DS(
                "CanonicalizationMethod",
                Algorithm="http://www.w3.org/2001/10/xml-exc-c14n#",
            ),
            DS(
                "SignatureMethod",
                Algorithm="http://www.w3.org/2001/04/xmldsig-more#rsa-sha256",
            ),
            DS(
                "Reference",
                DS("Transforms", *transforms),
                DS("DigestMethod", Algorithm="http://www.w3.org/2001/04/xmlenc#sha256"),
                DS("DigestValue"),
                URI=f"#{id}",
            ),
        ),
        DS("SignatureValue"),
    ]
    return DS("Signature", *args)


def create_service_provider(
    service_provider_id: str,
    organization_display_name: str,
    service_definitions: list,
    service_instances: list,
) -> Element:
    ns = namespaces["esc"]
    org_name_elements = create_language_elements(
        "OrganizationDisplayName", organization_display_name
    )

    args = [
        ESC("ServiceProviderID", service_provider_id),
        *org_name_elements,
        *service_definitions,
        *service_instances,
    ]
    kwargs = {f"{{{ns}}}IsPublic": "true"}
    return ESC("ServiceProvider", *args, **kwargs)


def create_service_definition(
    service_uuid,
    service_name,
    service_description,
    loa,
    entity_concerned_types_allowed,
    requested_attributes,
    makelaar_oin,
):

    service_name_elements = create_language_elements("ServiceName", service_name)
    service_description_elements = create_language_elements(
        "ServiceDescription", service_description
    )

    ns = namespaces["esc"]
    args = [
        ESC("ServiceUUID", service_uuid),
        *service_name_elements,
        *service_description_elements,
        SAML("AuthnContextClassRef", loa),
        ESC("HerkenningsmakelaarId", makelaar_oin),
    ]

    for entity in entity_concerned_types_allowed:
        assert isinstance(entity, dict)

        kwargs = {}
        set_number = entity.get("set_number")
        if set_number:
            kwargs["setNumber"] = set_number
        args.append(
            ESC("EntityConcernedTypesAllowed", entity["name"], **kwargs),
        )

    for requested_attribute in requested_attributes:
        if isinstance(requested_attribute, dict):
            ra_kwargs = {}
            if "required" in requested_attribute:
                ra_kwargs["isRequired"] = (
                    "true" if requested_attribute["required"] else "false"
                )

            ra_args = []
            if not "purpose_statements" in requested_attribute:
                ra_args += create_language_elements("PurposeStatement", service_name)
            else:
                ra_args += create_language_elements(
                    "PurposeStatement", requested_attribute["purpose_statements"]
                )

            ra_kwargs["Name"] = requested_attribute["name"]
            args.append(
                ESC("RequestedAttribute", *ra_args, **ra_kwargs),
            )
        else:
            args.append(
                ESC("RequestedAttribute", Name=requested_attribute),
            )

    kwargs = {f"{{{ns}}}IsPublic": "true"}
    return ESC("ServiceDefinition", *args, **kwargs)


def create_service_instance(
    service_id,
    service_uuid,
    instance_of_service,
    service_url,
    privacy_policy_url,
    herkenningsmakelaars_id,
    classifiers,
    key_descriptor,
):
    ns = namespaces["esc"]

    privacy_url_elements = create_language_elements(
        "PrivacyPolicyURL", privacy_policy_url
    )

    args = [
        ESC("ServiceID", service_id),
        ESC("ServiceUUID", service_uuid),
        ESC("InstanceOfService", instance_of_service),
        ESC("ServiceURL", service_url, **xml_nl_lang),
        *privacy_url_elements,
        ESC("HerkenningsmakelaarId", herkenningsmakelaars_id),
        ESC("SSOSupport", "false"),
        ESC("ServiceCertificate", key_descriptor),
    ]
    if len(classifiers) > 0:
        args.append(create_classifiers_element(classifiers))

    kwargs = {f"{{{ns}}}IsPublic": "true"}
    return ESC("ServiceInstance", *args, **kwargs)


def create_classifiers_element(classifiers: list) -> ElementMaker:
    classifiers_elements = []
    for classifier in classifiers:
        classifiers_elements.append(ESC("Classifier", classifier))
    return ESC("Classifiers", *classifiers_elements)


def create_key_descriptor(x509_certificate_content):
    x509_certificate = crypto.load_certificate(
        crypto.FILETYPE_PEM, x509_certificate_content
    )
    key_descriptor_cert = b64encode(
        crypto.dump_certificate(crypto.FILETYPE_ASN1, x509_certificate)
    ).decode("ascii")

    certificate = x509_certificate.to_cryptography()
    key_name = binascii.hexlify(
        certificate.fingerprint(certificate.signature_hash_algorithm)
    ).decode("ascii")

    args = [
        DS(
            "KeyInfo",
            DS("KeyName", key_name),
            DS("X509Data", DS("X509Certificate", key_descriptor_cert)),
        )
    ]
    kwargs = {"use": "encryption"}
    return MD("KeyDescriptor", *args, **kwargs)


def create_service_catalogus(conf, validate=True):
    """
    https://afsprakenstelsel.etoegang.nl/display/as/Service+catalog
    """
    x509_certificate_content = open(conf["cert_file"], "rb").read()

    sc_id = str(uuid4())
    service_provider_id = conf["oin"]
    organization_display_name = conf["organisation_name"]

    service_definitions = []
    service_instances = []
    for service in conf["services"]:
        key_descriptor = create_key_descriptor(x509_certificate_content)

        # https://afsprakenstelsel.etoegang.nl/display/as/ServiceUUID
        service_uuid = service["service_uuid"]
        service_name = service["service_name"]
        service_description = service["service_description"]
        # https://afsprakenstelsel.etoegang.nl/display/as/Level+of+assurance
        service_loa = service["service_loa"]
        # https://afsprakenstelsel.etoegang.nl/display/as/ServiceID
        service_id = "urn:etoegang:DV:{}:services:{}".format(
            conf["oin"], service["attribute_consuming_service_index"]
        )
        service_instance_uuid = service["service_instance_uuid"]

        service_url = service.get(
            "service_url",
        )
        privacy_policy_url = service.get(
            "privacy_policy_url",
        )
        herkenningsmakelaars_id = service.get(
            "herkenningsmakelaars_id",
        )
        entity_concerned_types_allowed = service.get("entity_concerned_types_allowed")
        requested_attributes = service.get("requested_attributes", [])
        classifiers = service.get("classifiers", [])

        service_definition = create_service_definition(
            service_uuid,
            service_name,
            service_description,
            service_loa,
            entity_concerned_types_allowed,
            requested_attributes,
            herkenningsmakelaars_id,
        )
        service_instance = create_service_instance(
            service_id,
            service_instance_uuid,
            service_uuid,
            service_url,
            privacy_policy_url,
            herkenningsmakelaars_id,
            classifiers,
            key_descriptor,
        )

        service_definitions.append(service_definition)
        service_instances.append(service_instance)

    signature = create_signature(sc_id)
    service_provider = create_service_provider(
        service_provider_id,
        organization_display_name,
        service_definitions,
        service_instances,
    )
    xml = create_service_catalogue(sc_id, timezone.now(), signature, service_provider)

    catalogus = tostring(xml, pretty_print=True, xml_declaration=True, encoding="utf-8")
    if validate:
        errors = validate_xml(BytesIO(catalogus), EHERKENNING_DS_XSD)
        assert errors is None, errors
    return catalogus


def get_metadata_eherkenning_requested_attributes(
    conf: dict, service_id: str
) -> List[dict]:
    # There needs to be a RequestedAttribute element where the name is the ServiceID
    # https://afsprakenstelsel.etoegang.nl/display/as/DV+metadata+for+HM
    requested_attributes = [{"name": service_id, "isRequired": False}]
    for requested_attribute in conf.get("requested_attributes", []):
        if isinstance(requested_attribute, dict):
            requested_attributes.append(
                {
                    "name": requested_attribute["name"],
                    "isRequired": requested_attribute["required"],
                }
            )
        else:
            requested_attributes.append(
                {
                    "name": requested_attribute,
                    "isRequired": True,
                }
            )

    return requested_attributes


def create_attribute_consuming_services(conf: dict) -> list:
    attribute_consuming_services = []

    for service in conf["services"]:
        service_id = "urn:etoegang:DV:{}:services:{}".format(
            conf["oin"], service["attribute_consuming_service_index"]
        )
        service_name = get_service_name(service)
        service_description = get_service_description(service)
        requested_attributes = get_metadata_eherkenning_requested_attributes(
            service, service_id
        )

        attribute_consuming_services.append(
            {
                "index": service["attribute_consuming_service_index"],
                "serviceName": service_name,
                "serviceDescription": service_description,
                "requestedAttributes": requested_attributes,
                "language": service.get("language", "nl"),
            }
        )
    return attribute_consuming_services


class eHerkenningClient(BaseSaml2Client):
    cache_key_prefix = "eherkenning"
    cache_timeout = 60 * 60  # 1 hour

    def __init__(self):
        conf = settings.EHERKENNING.copy()
        conf.setdefault("acs_path", reverse("eherkenning:acs"))

        super().__init__(conf)

    def write_metadata(self):
        """
        Write the dienstencatalogus, in addition to the saml2 metadata to the
        current working directory.

        :raises FileExistsError
        """
        super().write_metadata()

        service_catalogus = create_service_catalogus(settings.EHERKENNING)

        date_string = timezone.now().date().isoformat()
        dc_filename = f"eherkenning-dienstencatalogus-{date_string}.xml"
        dc_file = open(dc_filename, "xb")
        dc_file.write(service_catalogus)
        dc_file.close()

    def create_config_dict(self, conf):
        config_dict = super().create_config_dict(conf)

        attribute_consuming_services = create_attribute_consuming_services(conf)
        config_dict.update(
            {
                "sp": {
                    # Identifier of the SP entity  (must be a URI)
                    "entityId": conf["entity_id"],
                    # Specifies info about where and how the <AuthnResponse> message MUST be
                    # returned to the requester, in this case our SP.
                    "assertionConsumerService": {
                        "url": conf["base_url"] + conf["acs_path"],
                        "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Artifact",
                    },
                    "attributeConsumingServices": attribute_consuming_services,
                    "NameIDFormat": "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified",
                    "x509cert": open(conf["cert_file"], "r").read(),
                    "privateKey": open(conf["key_file"], "r").read(),
                    "privateKeyPassphrase": conf.get("key_passphrase", None),
                },
            }
        )
        return config_dict

    def create_config(self, config_dict):
        config_dict["security"].update(
            {
                # See comment in the python3-saml for in  OneLogin_Saml2_Response.validate_num_assertions (onelogin/saml2/response.py)
                # for why we need this option.
                "disableSignatureWrappingProtection": True,
                # For eHerkenning, if the Metadata file expires, we sent them an update. So
                # there is no need for an expiry date.
                "metadataValidUntil": "",
                "metadataCacheDuration": "",
                "requestedAuthnContextComparison": "minimum",
                "requestedAuthnContext": [
                    self.conf["services"][0]["service_loa"],
                ],
            }
        )
        return super().create_config(config_dict)

    def create_authn_request(
        self, request, return_to=None, attr_consuming_service_index=None, **kwargs
    ):
        return super().create_authn_request(
            request,
            return_to=return_to,
            force_authn=True,
            is_passive=False,
            set_nameid_policy=False,
            name_id_value_req=None,
            attr_consuming_service_index=attr_consuming_service_index,
        )



import urllib
from typing import List

from django.conf import settings
from django.core.cache import cache
from django.core.exceptions import ImproperlyConfigured
from django.utils import timezone

from onelogin.saml2.auth import OneLogin_Saml2_Auth
from onelogin.saml2.idp_metadata_parser import OneLogin_Saml2_IdPMetadataParser
from onelogin.saml2.settings import OneLogin_Saml2_Settings
from onelogin.saml2.utils import OneLogin_Saml2_ValidationError

from digid_eherkenning.utils import get_client_ip


def create_saml2_request(base_url, request):
    #
    # Because there might be proxying done before finally
    # getting to the Django server and SERVER_NAME and SERVER_PORT in request.META
    # might not be set correctly, instead, we hard-code these parameters
    # based on settings.
    #
    # X-Forwarded-For is also not an option, because it only forwards the
    # IP-Address.
    #
    parsed_url = urllib.parse.urlparse(base_url)
    return {
        "https": "on" if parsed_url.scheme == "https" else "off",
        "http_host": parsed_url.netloc,
        "script_name": request.META["PATH_INFO"],
        "server_port": parsed_url.port,
        "get_data": request.GET.copy(),
        "post_data": request.POST.copy(),
        # Uncomment if using ADFS as IdP, https://github.com/onelogin/python-saml/pull/144
        # 'lowercase_urlencoding': True,
        "query_string": request.META["QUERY_STRING"],
    }


def get_service_name(conf: dict) -> str:
    _service_name = conf.get("service_name", "")
    return _service_name["en"] if isinstance(_service_name, dict) else _service_name


def get_service_description(conf: dict) -> str:
    _service_description = conf.get("service_description", "")
    return (
        _service_description["en"]
        if isinstance(_service_description, dict)
        else _service_description
    )


def get_requested_attributes(conf: dict) -> List[dict]:
    # There needs to be a RequestedAttribute element where the name is the ServiceID
    # https://afsprakenstelsel.etoegang.nl/display/as/DV+metadata+for+HM
    requested_attributes = []
    for requested_attribute in conf.get("requested_attributes", []):
        if isinstance(requested_attribute, dict):
            requested_attributes.append(requested_attribute)
        else:
            requested_attributes.append(
                {
                    "name": requested_attribute,
                    "required": True,
                }
            )

    return requested_attributes


class BaseSaml2Client:
    cache_key_prefix = "saml2_"
    cache_timeout = 60 * 60  # 1 hour

    def __init__(self, conf):
        self.conf = conf
        self.saml2_settings = self.create_config(
            self.create_config_dict(conf=self.conf),
        )

        self.authn_storage = AuthnRequestStorage(
            self.cache_key_prefix, self.cache_timeout
        )

    def create_metadata(self):
        return self.saml2_settings.get_sp_metadata()

    def get_saml_metadata_path(self):
        """
        File is written to the current working directory by default.
        """
        date_string = timezone.now().date().isoformat()
        return f"{self.cache_key_prefix}-metadata-{date_string}.xml"

    def write_metadata(self):
        """
        Write SAML metadata to the path specified by get_saml_metadata_path.

        :raises FileExistsError
        """
        metadata_content = self.create_metadata()
        metadata_file = open(self.get_saml_metadata_path(), "xb")
        metadata_file.write(metadata_content)
        metadata_file.close()

    def create_authn_request(
        self, request, return_to=None, attr_consuming_service_index=None, **kwargs
    ):
        saml2_request = create_saml2_request(self.conf["base_url"], request)
        saml2_auth = OneLogin_Saml2_Auth(
            saml2_request, old_settings=self.saml2_settings, custom_base_path=None
        )
        url, parameters = saml2_auth.login_post(
            return_to=return_to,
            attr_consuming_service_index=attr_consuming_service_index,
            **kwargs,
        )

        # Save the request ID so we can verify that we've sent
        # it when we receive the Artifact/ACS response.
        request_id = saml2_auth.get_last_request_id()
        self.authn_storage.store(request_id, get_client_ip(request))

        return url, parameters

    def artifact_resolve(self, request, saml_art):
        saml2_request = create_saml2_request(self.conf["base_url"], request)
        saml2_auth = OneLogin_Saml2_Auth(
            saml2_request, old_settings=self.saml2_settings, custom_base_path=None
        )
        response = saml2_auth.artifact_resolve(saml_art)

        self.verify_saml2_response(response, get_client_ip(request))

        return response

    def handle_post_response(self, request):
        saml2_request = create_saml2_request(self.conf["base_url"], request)

        saml2_auth = OneLogin_Saml2_Auth(
            saml2_request, old_settings=self.saml2_settings, custom_base_path=None
        )

        response = saml2_auth.post_response()

        self.verify_saml2_response(response, get_client_ip(request))

        return response

    def verify_saml2_response(self, response, client_ip_address):
        #
        # SAMLProf: 4.1.4.2 <Response> Usage
        #
        # If the containing message is in response to an <AuthnRequest>,
        # then the InResponseTo attribute MUST match the request's ID
        #
        in_response_to = response.get_in_response_to()
        authn_request = self.authn_storage.get(in_response_to)
        if authn_request is None:
            raise OneLogin_Saml2_ValidationError(
                f"The InResponseTo of the Response: {in_response_to}, is not a request id"
                "found in the request cache",
                OneLogin_Saml2_ValidationError.WRONG_INRESPONSETO,
            )

        #
        # This is not a mandatory check by the SAML specification. But seems
        # like a good idea to guard against various attacks.
        #
        authn_ip_address = authn_request["client_ip_address"]
        if authn_ip_address != client_ip_address:
            raise OneLogin_Saml2_ValidationError(
                f"A different IP address ({authn_ip_address})"
                f"was used when retrieving the AuthNRequest then for retrieving"
                f" the request to the ACS ({client_ip_address}).",
                OneLogin_Saml2_ValidationError.WRONG_INRESPONSETO,
            )

    def create_config(self, config_dict):
        """
        Convert to the format expected by the OneLogin SAML2 library.
        """
        return OneLogin_Saml2_Settings(config_dict, custom_base_path=None)

    def create_config_dict(self, conf):
        """
        Convert the settings specified in conf, which has the following format

        base_url: URL which is prefixed before any URL used by the SP we set up.
        entity_id: SAML2 Entity id of SP we set up.
        key_file: path on disk of private key in PEM format
        cert_file: path on disk of certificate in PEM format.

        metadata_file: path on disk for metadata file which specifies the IDP.
        service_entity_id: entity id used to find settings of IDP in metadata file.
        attribute_consuming_service_index
        service_name: Name of SP service we set up. This is used in metadata generation.
        requested_attributes: List of attributes which should be returned by the IDP.
        """
        try:
            metadata_content = open(conf["metadata_file"], "r").read()
        except FileNotFoundError:
            raise ImproperlyConfigured(
                f"The file: {conf['metadata_file']} could not be found. Please "
                "specify an existing metadata in the conf['metadata_file'] setting."
            )

        idp_settings = OneLogin_Saml2_IdPMetadataParser.parse(
            metadata_content, entity_id=conf["service_entity_id"]
        )["idp"]

        service_name = get_service_name(conf)
        service_description = get_service_description(conf)
        requested_attributes = get_requested_attributes(conf)

        setting_dict = {
            "strict": True,
            "security": {
                "signMetadata": True,
                "authnRequestsSigned": True,
                "wantAssertionsEncrypted": conf.get("want_assertions_encrypted", False),
                "wantAssertionsSigned": conf.get("want_assertions_signed", False),
                "soapClientKey": conf["key_file"],
                "soapClientCert": conf["cert_file"],
                "soapClientPassphrase": conf.get("key_passphrase", None),
                "signatureAlgorithm": conf.get("signature_algorithm"),
                "digestAlgorithm": conf.get("digest_algorithm"),
            },
            "debug": settings.DEBUG,
            # Service Provider Data that we are deploying.
            "sp": {
                # Identifier of the SP entity  (must be a URI)
                "entityId": conf["entity_id"],
                # Specifies info about where and how the <AuthnResponse> message MUST be
                # returned to the requester, in this case our SP.
                "assertionConsumerService": {
                    "url": conf["base_url"] + conf["acs_path"],
                    "binding": "urn:oasis:names:tc:SAML:2.0:bindings:HTTP-Artifact",
                },
                # If you need to specify requested attributes, set a
                # attributeConsumingService per service. nameFormat, attributeValue and
                # friendlyName can be omitted
                "attributeConsumingService": {
                    "index": conf.get("attribute_consuming_service_index", "1"),
                    "serviceName": service_name,
                    "serviceDescription": service_description,
                    "requestedAttributes": [
                        {
                            "name": attr["name"],
                            "isRequired": True if attr["required"] else False,
                        }
                        for attr in requested_attributes
                    ],
                },
                "NameIDFormat": "urn:oasis:names:tc:SAML:1.1:nameid-format:unspecified",
                "x509cert": open(conf["cert_file"], "r").read(),
                "privateKey": open(conf["key_file"], "r").read(),
                "privateKeyPassphrase": conf.get("key_passphrase", None),
            },
            "idp": idp_settings,
        }

        telephone = conf.get("technical_contact_person_telephone")
        email = conf.get("technical_contact_person_email")
        if telephone or email:
            setting_dict["contactPerson"] = {
                "technical": {"telephoneNumber": telephone, "emailAddress": email}
            }

        organisation = conf.get("organization")
        if organisation:
            setting_dict["organization"] = organisation

        return setting_dict


class AuthnRequestStorage:
    def __init__(self, cache_key_prefix, cache_timeout):
        self.cache_key_prefix = cache_key_prefix
        self.cache_timeout = cache_timeout

    def get_cache_key(self, request_id):
        return f"{self.cache_key_prefix}_{request_id}"

    def store(self, request_id, client_ip_address):
        """
        Save the request id and the ip address of the client in the cache.
        We use this later to check if this match if the user retrieves
        the response.
        """

        cache_key = self.get_cache_key(request_id)
        cache_value = {
            "current_time": timezone.now(),
            "client_ip_address": client_ip_address,
        }
        cache.set(cache_key, cache_value, self.cache_timeout)

    def get(self, request_id):
        cache_key = self.get_cache_key(request_id)
        cached_value = cache.get(cache_key)
        return cached_value
